import 'dart:async';
import 'dart:convert';
import 'package:dio/dio.dart';

import './log.dart';

//import 'dart:io';
//import 'package:flutter_secure_storage/flutter_secure_storage.dart' as Sec;


class AuthInfo {
	String token;
	int tokenExpire;

	AuthInfo(this.token, this.tokenExpire);
}

class DipClient {
	static String API_URL_BASE = 'https://api.identity-registry.com/';

	static String ENV_PRD = 'p';
	static String ENV_INT= 'i';
	static String ENV_DEV = 'd';

	BaseOptions? _baseOptions;

	String? _userName;
	String? _password;
	String? _token;
	int? _tokenExpire;

	DipClient(envKey) {
		_baseOptions = BaseOptions(
			baseUrl: API_URL_BASE + envKey,
			connectTimeout: 5000,
			receiveTimeout: 3000,
		);
	}

	AuthInfo? getAuthInfo() {
		if (_token != null && _tokenExpire != null) {
			return new AuthInfo(_token!, _tokenExpire!);
		}
		return null;
	}

	void clearAuthInfo() {
		_userName = null;
		_password = null;
		_token = null;
		_tokenExpire = null;
	}


	//-----------------------------------------------------------------------------------
	//
	// Internal implementation
	//
	//-----------------------------------------------------------------------------------

	Future<String> _getToken() async {
		if (_token != null) {
			var now = DateTime.now().millisecondsSinceEpoch;
			if (now <= _tokenExpire! - 60000) {
				// if now is earlier than one minute from expire, use the cached token
				return _token!;
			}

			// auto refresh token
			//
			if (_userName != null && _password != null) {
				await authenticate(userName: _userName!, password: _password!);
				return _token!;
			}
		}
		return "";
	}

	Future _invoke(String url, Map params, [bool authRequired = false]) async {
		Dio dio = Dio(_baseOptions);

		Map<String,dynamic> headers = {};

		if (authRequired && _token != null) {
			headers['Authorization'] = 'BEARER ' + await _getToken();
		}

		Response resp = await dio.post(url, data: params, options: Options(headers:headers)); 
		return jsonDecode(resp.data);
	}


	//-----------------------------------------------------------------------------------
	//
	// User Manipulation
	//
	//-----------------------------------------------------------------------------------

	Future createUser({
		required String userName,
		required String password,
		required String name,
		String? mobile,
		String? mobileVerification
	}) async {
		var url = '/create-user';
		var params = {
			'userName': userName,
			'password': password,
			'name': name,
		};

		if (mobile != null) {
			params['mobile'] = mobile;
		}
		if (mobileVerification != null) {
			params['mobileVerification'] = mobileVerification;
		}

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	Future updateUser({
		required String userName,
		required List updates
	}) async {
		var url = '/update-user';
		var params = {
			'userName': userName,
			'updates': updates,
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	Future getUser({
		required String userName
	}) async {
		var url = '/get-user';
		var params = {
			'userName': userName,
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		return resp['user'];
	}

	Future deleteUser({
		required String userName,
	}) async {
		var url = '/delete-user';
		var params = {
			'userName': userName,
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}


	Future authenticate({
		required String userName,
		required String password
	}) async {
		var url = '/authenticate';
		var params = {
			'userName': userName,
			'password': password
		};

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];

		_userName = userName;
		_password = password;
		_token = resp['authResult']['token'];
		_tokenExpire = resp['authResult']['tokenExpire'];
	}

	Future authNaverMobile({
		required String accessToken
	}) async {
		var url = '/auth-naver-mobile';
		var params = {
			'accessToken': accessToken,
		};

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];

		_userName = resp['userName'];
		_token = resp['authResult']['token'];
		_tokenExpire = resp['authResult']['tokenExpire'];
	}

	Future authKakaoMobile({
		required String accessToken
	}) async {
		var url = '/auth-kakao-mobile';
		var params = {
			'accessToken': accessToken,
		};

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];

		_userName = resp['userName'];
		_token = resp['authResult']['token'];
		_tokenExpire = resp['authResult']['tokenExpire'];
	}

	Future authAppleMobile({
		required String accessToken
	}) async {
		var url = '/auth-apple-mobile';
		var params = {
			'accessToken': accessToken,
		};

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];

		_userName = resp['userName'];
		_token = resp['authResult']['token'];
		_tokenExpire = resp['authResult']['tokenExpire'];
	}

	Future checkUserName({
		required String userName
	}) async {
		var url = '/check-username';
		var params = {
			'userName': userName,
		};

		var resp = await _invoke(url, params);

		if (resp['result'] == 'OK') {
			return true;
		}
		else if (resp['result'] == 'NOT_ALLOWED') {
			return false;
		}
		else {
			throw resp['result'];
		}
	}

	Future findUserName({
		required String mobile,
		required String name,
		String? userNamePrefix,
		String? smsHeader,
	}) async {
		var url = '/find-username';
		var params = {
			'mobile': mobile,
			'name': name,
		};

		if (userNamePrefix != null) {
			params['userNamePrefix'] = userNamePrefix;
		}
		if (smsHeader != null) {
			params['smsHeader'] = smsHeader;
		}

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	Future requestMobileVerification({
		required String mobile,
		String? smsHeader,
		String? userName
	}) async {
		var url = '/request-mobile-verification';
		var params = {
			'mobile': mobile
		};

		if (smsHeader != null) {
			params['smsHeader'] = smsHeader;
		}

		if (userName != null) {
			params['userName'] = userName;
		}

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	Future submitMobileVerification({
		required String mobile,
		required String verificationCode
	}) async {
		var url = '/submit-mobile-verification';
		var params = {
			'mobile': mobile,
			'verificationCode': verificationCode
		};

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	Future resetPassword({
		required String userName,
		required String mobile,
		required String mobileVerification,
		required String newPassword
	}) async {
		var url = '/reset-password';
		var params = {
			'userName': userName,
			'mobile': mobile,
			'mobileVerification': mobileVerification,
			'newPassword': newPassword
		};

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];
	}


	//-----------------------------------------------------------------------------------
	//
	// Resource Manipulation
	//
	//-----------------------------------------------------------------------------------

	Future<List<Map<String,dynamic>>> addResources({
		String? userName,
		required List<Map<String,dynamic>> resources
	}) async {
		var url = '/add-resources';
		var params = <String,dynamic>{
			'resources': resources
		};

		if (userName != null) {
			params['userName'] = userName;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		List<Map<String,dynamic>> resourceUrls = [];
		resp['resourceUrls'].forEach((dynamic e) => resourceUrls.add(e as Map<String,dynamic>));

		return resourceUrls;
	}

	Future<List<Map<String,dynamic>>> getResources({
		required List<String> resourceIds
	}) async {
		var url = '/get-resources';
		var params = {
			"resourceIds": resourceIds
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		List<Map<String,dynamic>> resources = [];
		resp['resources'].forEach((dynamic e) => resources.add(e as Map<String,dynamic>));

		return resources;
	}

	Future<List<Map<String,dynamic>>> getResourceStatus({
		required List<String> resourceIds
	}) async {
		var url = '/get-resource-status';
		var params = {
			"resourceIds": resourceIds
		};

		var resp = await _invoke(url, params, true);

		List<Map<String,dynamic>> resources = [];
		resp['resources'].forEach((dynamic e) => resources.add(e as Map<String,dynamic>));

		return resources;
	}

	Future<List<Map<String,dynamic>>> getResourcesByTag({
		required String tag
	}) async {
		var url = '/get-resources-by-tag';
		var params = {
			"tag": tag
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		List<Map<String,dynamic>> resources = [];
		resp['resources'].forEach((dynamic e) => resources.add(e as Map<String,dynamic>));

		return resources;
	}


	//-----------------------------------------------------------------------------------
	//
	// Identity Manipulation
	//
	//-----------------------------------------------------------------------------------

	Future createIdentity({
		required String category,
		String? identityOwner,
		String? redirectIdentity,
		String? redirectUrl,
		Map<String,dynamic>? acl,
		Map<String,dynamic>? attrs,
		Map<String,dynamic>? dattrs,
		Map<String,dynamic>? resources,
		dynamic? accessNotifyPolicy,
		bool? isTemporary,
		List<String>? extensions
	}) async {
		var url = '/create-identity';
		var params = <String,dynamic>{
			'category': category
		};

		if (identityOwner != null) {
			params['identityOwner'] = identityOwner;
		}
		if (redirectIdentity != null) {
			params['redirectIdentity'] = redirectIdentity;
		}
		if (redirectUrl != null) {
			params['redirectUrl'] = redirectUrl;
		}
		if (acl != null) {
			params['acl'] = acl;
		}
		if (attrs != null) {
			params['attrs'] = attrs;
		}
		if (dattrs != null) {
			params['dattrs'] = dattrs;
		}
		if (resources != null) {
			params['resources'] = resources;
		}
		if (isTemporary != null) {
			params['isTemporary'] = isTemporary;
		}
		if (accessNotifyPolicy != null) {
			params['accessNotifyPolicy'] = accessNotifyPolicy;
		}
		if (extensions != null) {
			params['extensions'] = extensions;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		return resp['id'] as String;
	}

	Future getIdentities({
		required List<String> identityIds
	}) async {
		var url = '/get-identities';
		var params = {
			'identityIds': identityIds,
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
		return resp['identities'];
	}

	Future getIdentitiesByUser({
		required String userName,
		num? tFrom,
		num? tUntil,
		bool? isTemporary
	}) async {
		var url = '/get-identities-by-user';
		var params = <String,dynamic>{
			'userName': userName
		};

		if (tFrom != null) {
			params['tFrom'] = tFrom;
		}
		if (tUntil != null) {
			params['tUntil'] = tUntil;
		}
		if (isTemporary != null) {
			params['isTemporary'] = isTemporary;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
		return resp['identities'];
	}

	Future updateIdentity({
		required String identityId,
		required List updates
	}) async {
		var url = '/update-identity';
		var params = {
			'identityId': identityId,
			'updates': updates
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	Future deleteIdentity({
		required String identityId
	}) async {
		var url = '/delete-identity';
		var params = {
			'identityId': identityId
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	Future accessIdentity({
		required String identityId,
		Map<String,dynamic>? accessParams,
		bool? returnValues
	}) async {
		var url = '/access-identity';
		var params = <String,dynamic>{
			'identityId': identityId
		};

		if (accessParams != null) {
			params['accessParams'] = accessParams;
		}
		if (returnValues != null) {
			params['returnValues'] = returnValues;
		}

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];

		return {
			'attrs': resp['attrs'],
			'dattrs': resp['dattrs'],
			'resources': resp['resources'],
		};
	}


	//-----------------------------------------------------------------------------------
	//
	// Group Manipulation
	//
	//-----------------------------------------------------------------------------------

	/// Creates a group that can hold various kinds of entities such as identity and
	/// identity template
	///
	Future createGroup({
		required String category,
		required String name,
		String? groupOwner,
		Map<String,dynamic>? attrs,
		bool? useInlineMembers = true,
	}) async {
		var url = '/create-group';
		var params = <String,dynamic>{
			"category": category,
			"name": name,
		};

		if (groupOwner != null) {
			params['groupOwner'] = groupOwner;
		}
		if (attrs != null) {
			params['attrs'] = attrs;
		}
		if (useInlineMembers != null) {
			params['useInlineMembers'] = useInlineMembers;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	/// Deletes a group
	///
	Future deleteGroup({
		required String groupCategory,
		required String groupName,
	}) async {
		var url = '/delete-group';
		var params = <String,dynamic>{
			"groupCategory": groupCategory,
			"groupName": groupName,
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	Future getGroups({
		required List<Map> groupKeys
	}) async {
		var url = '/get-groups';
		var params = {
			"groupKeys": groupKeys
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		return <String,dynamic>{
			'groups': resp['groups'],
		};
	}


	/// Find groups that matches to the given criteria. 'name' parameter is for exact match and
	/// can result in only one group returned but 'namePrefix' parameter can result in multiple
	/// groups returned. 'name' and 'namePrefix' parameters are mutually exclusive and only either
	/// of them can be specified (not both).
	///
	Future findGroups({
		required String category,
		String? name,
		String? namePrefix,
		dynamic? lastKey
	}) async {
		var url = '/find-groups';
		var params = {
			"category": category
		};

		if (name != null) {
			params['name'] = name;
		}
		if (namePrefix != null) {
			params['namePrefix'] = namePrefix;
		}
		if (lastKey != null) {
			params['lastKey'] = lastKey;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		return <String,dynamic>{
			'groups': resp['groups'],
			'lastKey': resp['lastKey']
		};
	}

	/// Adds a member to the group. Possible memberType (at the time of this writing) are:
	///
	/// 'id': identity
	/// 'it': identity template
	///
	Future addToGroup({
		required String groupCategory,
		required String groupName,
		required String memberType,
		required String memberId,
		required String memberName,
		Map<String,dynamic>? groupCreateParams,
		Map<String,dynamic>? joinAttrs,
	}) async {
		var url = '/add-to-group';
		var params = <String,dynamic>{
			"groupCategory": groupCategory,
			"groupName": groupName,
			"memberType": memberType,
			"memberId": memberId,
			"memberName": memberName,
		};

		if (groupCreateParams != null) {
			params['groupCreateParams'] = groupCreateParams;
		}

		if (joinAttrs != null) {
			params["joinAttrs"] = joinAttrs;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	/// Removes a member from the group. Possible memberType (at the time of this writing) are:
	///
	/// 'id': identity
	/// 'it': identity template
	///
	Future removeFromGroup({
		required String groupCategory,
		required String groupName,
		required String memberType,
		required String memberName
	}) async {
		var url = '/remove-from-group';
		var params = <String,dynamic>{
			"groupCategory": groupCategory,
			"groupName": groupName,
			"memberType": memberType,
			"memberName": memberName,
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	/// Returns all members of the group. If there are too many members to be returned from
	/// a single call of this API, 'lastKey' is returned in addition to the 'members' and
	/// the 'lastKey' can be used for the next call to retrieve more members.
	///
	Future getGroupMembers({
		required String groupCategory,
		required String groupName,
		dynamic? lastKey
	}) async {
		var url = '/get-group-members';
		var params = <String,dynamic>{
			"groupCategory": groupCategory,
			"groupName": groupName
		};

		if (lastKey != null) {
			params['lastKey'] = lastKey;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		return <String,dynamic>{
			'members': resp['members'],
			'lastKey': resp['lastKey']
		};
	}


	//-----------------------------------------------------------------------------------
	//
	// Link Identities
	//
	//-----------------------------------------------------------------------------------

	/// return redirect information (redirectIdentity, redirectUrl) about identity if any
	///
	Future getRedirectInfo({
		required String identityId
	}) async {
		var url = '/get-redirect-info';
		var params = <String,dynamic>{
			'identityId': identityId
		};

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];

		return {
			'redirectIdentity': resp['redirectIdentity'],
			'redirectUrl': resp['redirectUrl'],
		};
	}

	/// link an identity to another identity so that if the first identity is accessed, the
	/// second identity (linked one; targetIdentityId) is presented instead.
	///
	Future linkIdentity({
		required String identityId,
		required String targetIdentityId,
		Map<String,dynamic>? linkAttrs,
	}) async {
		var url = '/link-identity';
		var params = <String,dynamic>{
			'identityId': identityId,
			'targetIdentityId': targetIdentityId
		};

		if (linkAttrs != null) {
			params["linkAttrs"] = linkAttrs;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	/// Reverse the effect of linkIdentity API
	///
	Future unlinkIdentity({
		required String identityId,
		required String targetIdentityId,
	}) async {
		var url = '/unlink-identity';
		var params = <String,dynamic>{
			"identityId": identityId,
			"targetIdentityId": targetIdentityId,
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	/// Unlink all identities linked to the given identity
	///
	Future unlinkAll({
		required String targetIdentityId,
	}) async {
		var url = '/unlink-all';
		var params = <String,dynamic>{
			"targetIdentityId": targetIdentityId,
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];
	}

	/// Returns all linked identities of the given target identity. If there are too many identities
	/// to be returned from a single call of this API, 'lastKey' is returned in addition to the
	/// 'identities' and the 'lastKey' can be used for the next call to retrieve more identities.
	///
	Future getLinkedIdentities({
		required String targetIdentityId,
		dynamic? lastKey
	}) async {
		var url = '/get-linked-identities';
		var params = <String,dynamic>{
			"targetIdentityId": targetIdentityId
		};

		if (lastKey != null) {
			params['lastKey'] = lastKey;
		}

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		return <String,dynamic>{
			'identities': resp['identities'],
			'lastKey': resp['lastKey']
		};
	}

	/// Returns the numbers of linked identities for given target identity IDs
	///
	Future getLinkedIdentityNums({
		required List<String> targetIdentityIds,
	}) async {
		var url = '/get-linked-identity-nums';
		var params = <String,dynamic>{
			"targetIdentityIds": targetIdentityIds
		};

		var resp = await _invoke(url, params, true);

		if (resp['result'] != 'OK') throw resp['result'];

		return resp['identityNums'];
	}


	//-----------------------------------------------------------------------------------
	//
	// Misc.
	//
	//-----------------------------------------------------------------------------------

	/// Gets a sequence number for the given sequence name. It's guaranteed that the returned
	/// number of unique across all invocations of this API for the given sequence name. However,
	/// multiple clients may invoke this API for the same sequence name, so invoking this API
	/// multiple times may NOT result in consecutive numbers.
	///
	Future getSeqNum({
		required String seqName
	}) async {
		var url = '/get-seq-num';
		var params = {
			'seqName': seqName
		};

		var resp = await _invoke(url, params);

		if (resp['result'] != 'OK') throw resp['result'];

		return resp['seqNum'] as int;
	}
}
