import 'dart:async';
import 'package:dio/dio.dart';

import './UploadFile.dart';
import './log.dart';

class UploadException implements Exception {
	int value;
	UploadException(this.value);
}

class MultiUploader {
	static const int FileStatus_NotUploaded = 0;
	static const int FileStatus_Uploading = 1;
	static const int FileStatus_Uploaded = 2;
	static const int FileStatus_Failed = 3;
	static const int FileStatus_Canceled = 4;

	List<Map<String,dynamic>> files = [];

	Map<String,dynamic> config = {
		'maxConcurrency': 5,
		'maxRetry': 3,
		'monitorInterval': 3000,
		'cancelThreshold': 0
	};

	bool isUploading = false;
	bool isAborted = false;

	MultiUploader(List<UploadFile> fileInfos, Map<String,dynamic>? config) {
		for (var i = 0; i < fileInfos.length; i++) {
			var fileObj = {
				'url': fileInfos[i].url,
				'file': fileInfos[i].file,
				'contentType': fileInfos[i].contentType,
				'fileName': fileInfos[i].fileName,
				'fileSize': fileInfos[i].fileSize,
				'status': FileStatus_NotUploaded,
				'loadedSize': 0,
				'lastLoadedSize': 0,
				'retryCount': 0 
			};
			Log(fileObj.toString());

			// sort this.files in the decreasing order of fileSize
			// insertion sort algorithm
			//
			var inserted = false;
			for (var j = 0; j < this.files.length; j++) {
				if (this.files[j]['fileSize'] < fileObj['fileSize']) {
					this.files.insert(j, fileObj);
					inserted = true;
					break;
				}
			}

			if (!inserted) {
				this.files.add(fileObj);
			}
		}

		if (config != null) {
			if (config['maxConcurrency'] != null) this.config['maxConcurrency'] = config['maxConcurrency'];
			if (config['maxRetry'] != null) this.config['maxRetry'] = config['maxRetry'];
			if (config['monitorInterval'] != null) this.config['monitorInterval'] = config['monitorInterval'];
			if (config['cancelThreshold'] != null) this.config['cancelThreshold'] = config['cancelThreshold'];
		}
	}

/*
	simulateUpload(resolve, reject, idx) {
		if (idx != 10) {
			Log("COMPLETED: " + idx);
			resolve(idx);
		}
		else {
			Log("FAILED: " + idx);
			reject(idx);
		}
	}
*/

	Future<int> uploadFile(int idx) async {
		Log("uploadFile: idx = " + idx.toString());

		this.files[idx]['status'] = FileStatus_Uploading;
		this.files[idx]['loadedSize'] = 0;

		//var uploadTime = Math.floor(5000 * Math.random()) + 3000;
		//setTimeout(this.simulateUpload.bind(this), uploadTime, resolve, reject, idx);

		this.files[idx]['cancelToken'] = CancelToken();

		Dio dio = Dio();
		var inputStream = this.files[idx]['file'].openRead();

		try {
			await dio.put(
				this.files[idx]['url'],
				data: this.files[idx]['file'].openRead(),
				//maxContentLength: Infinity,
				//maxBodyLength: Infinity,
				options: Options(
					contentType: this.files[idx]['contentType'],
					headers: {
						'Accept': "*/*",
						'Content-Length': this.files[idx]['fileSize'],
						//'Connection': 'keep-alive',
						//'User-Agent': 'ClinicPlush'
					}
				),
				cancelToken: this.files[idx]['cancelToken'],
				onSendProgress: (int sent, int total) {
					//Log("upload progress: idx = " + idx.toString() +
					//	", sent = " + sent.toString() + ", total = " + total.toString());

					// onSendProgress being called does not necessarily means the file
					// has been successfully saved on the server. The server may reject
					// the upload after all for various reasons.
					//
					this.files[idx]['loadedSize'] = sent;
				}
			);
			Log("upload succeeded");
			return idx;
		}
		on DioError catch (e) {
			Log(e.toString());
			if (CancelToken.isCancel(e)) {
				Log("upload canceled");
				this.files[idx]['status'] = FileStatus_Canceled;
				throw new UploadException(-1);
			}
			else {
				Log("upload failed");
				throw new UploadException(idx);
			}
		}
	}

	void monitor(Timer t) {
		//Log("monitor");

		if (!this.isUploading) {
			t.cancel();
		}

		for (var i = 0; i < this.files.length; i++) {
			if (this.files[i]['loadedSize'] < this.files[i]['fileSize'] &&
				this.files[i]['loadedSize'] - this.files[i]['lastLoadedSize'] < this.config['cancelThreshold']) {
				Log("upload is too slow: file = " + this.files[i]['fileName']);
				// [TODO] shall we cancel this upload?
				//this.abortEntireUpload();
			}
			this.files[i]['lastLoadedSize'] = this.files[i]['loadedSize'];
		}
	}

	int setup(List<Future> futures, List<int> idxForFutures) {
		var i = 0;

		while (i < this.files.length && i < this.config['maxConcurrency']) {
			futures.add(this.uploadFile(i));
			idxForFutures.add(i);
			i++;
		}
		return i;
	}

	void abortEntireUpload() {
		Log("aborting entire upload");
		for (var i = 0; i < this.files.length; i++) {
			if (this.files[i]['status'] == FileStatus_Uploading) {
				this.files[i]['cancelToken'].cancel('canceled');
				this.files[i]['status'] = FileStatus_Canceled;
			}
		}
		this.isAborted = true;
		Log("entire upload aborted");
	}

	Future<bool> upload() async {
		if (this.isUploading) {
			Log("already uploading");
			return false;
		}
		this.isUploading = true;
		this.isAborted = false;

		List<Future> futures = <Future>[];
		List<int> idxForFutures = <int>[];
		int idx = this.setup(futures, idxForFutures);

		var monitorInterval = Duration(milliseconds: this.config['monitorInterval']);
		var monitorTimer = Timer.periodic(monitorInterval, this.monitor);

		while (futures.length > 0 && !this.isAborted) {
			try {
				//Log("futures.length = " + futures.length.toString());

				var completedIdx = await Future.any(futures);
				Log("completedIdx = " + completedIdx.toString());

				this.files[completedIdx]['status'] = FileStatus_Uploaded;
				this.files[completedIdx]['loadedSize'] = this.files[completedIdx]['fileSize'];

				for (var i = 0; i < idxForFutures.length; i++) {
					if (idxForFutures[i] == completedIdx) {
						if (idx < this.files.length) {
							idxForFutures[i] = idx;
							futures[i] = this.uploadFile(idx);
							idx++;
						}
						else {
							idxForFutures.removeAt(i);
							futures.removeAt(i);
						}
						break;
					}
				}
			}
			on UploadException catch (e) {
				//Log("e = " + e.toString());
				if (e.value == -1) break;

				var failedIdx = e.value;

				// retrying...
				for (var i = 0; i < idxForFutures.length; i++) {
					if (idxForFutures[i] == failedIdx) {
						Log("failedIdx = " + failedIdx.toString() + ", retryCount = " +
							this.files[failedIdx]['retryCount'].toString());

						if (this.files[failedIdx]['retryCount'] < this.config['maxRetry']) {
							futures[i] = this.uploadFile(failedIdx);
							this.files[failedIdx]['retryCount']++;
						}
						else {
							idxForFutures.removeAt(i);
							futures.removeAt(i);
							this.files[failedIdx]['status'] = FileStatus_Failed;

							this.abortEntireUpload();
						}
						break;
					}
				}
			}
		}

		this.isUploading = false;

		// debug output
		//
		for (var i = 0; i < this.files.length; i++) {
			Log("idx: " + i.toString() +
				", status: " + this.files[i]['status'].toString() +
				", fileSize: " + this.files[i]['fileSize'].toString() +
				", loadedSize: " + this.files[i]['loadedSize'].toString() +
				", retryCount: " + this.files[i]['retryCount'].toString());
		}
		return !this.isAborted;
	}

	num getProgress() {
		num totalFileSize = 0;
		num totalLoadedSize = 0;
		for (var i = 0; i < this.files.length; i++) {
			totalFileSize += this.files[i]['fileSize'];
			totalLoadedSize += this.files[i]['loadedSize'];
		}
		return totalLoadedSize / totalFileSize;
	}
}
